//! Authentication for the API.
//!
//! Authentication is done using tokens that are generated by the
//! server. The SHA-256 hash of a token is stored in the
//! database. Since tokens are randomly generated by the server, they
//! are unique and therefore no salt is required.
//!
//! Tokens can be acquired by authenticating with one of the
//! configured OAuth providers. Part of the OAuth flow is implemented
//! in the [`frontend`](crate::frontend) module, and the rest is
//! implemented here.

use crate::database::Database;
use axum::extract::{FromRequest, RequestParts};
use http::header;
use serde::Deserialize;
use std::sync::Arc;
use tracing::{error, info};

/// Configuration for an OAuth provider.
#[derive(Deserialize)]
pub struct Provider {
    /// The name of the provider, to be displayed on the login page.
    pub name: String,

    /// The base URL of the auth endpoint, including a trailing slash.
    #[serde(with = "crate::de::url")]
    pub url: url::Url,

    /// The URL to which the client should be redirected. This must
    /// match the URL configured when registering an application with
    /// an OAuth provider.
    pub redirect_url: String,

    /// Client ID generated by an OAuth provided.
    pub client_id: String,

    /// Client secret generated by an OAuth provider.
    pub client_secret: String,

    /// Optional scopes to request access to. Defaults to an empty
    /// string. Some OAuth providers have default scopes which provide
    /// enough access to retrieve a user's username, which is all we
    /// need.
    #[serde(default)]
    pub scope: String,

    /// The API endpoint to query to retrieve basic user information.
    ///
    /// For example, GitHub returns user info by querying the
    /// `https://api.github.com/user` endpoint.
    #[serde(with = "crate::de::url")]
    pub user_url: url::Url,

    /// The path to extract from a JSON object returned from an API's
    /// user endpoint. This is used as the argument to
    /// [`serde_json::Value::pointer()`]; see that documentation for
    /// details of the syntax. For example, if the provider's user
    /// endpoint returns the following JSON, then the path would be
    /// `/user`:
    ///
    /// ```json
    /// {
    ///   "user": "myusername"
    /// }
    /// ```
    pub user_path: String,

    /// An optional user agent header. Some providers, such as GitHub,
    /// require this to be set.
    pub user_agent: Option<String>,
}

/// An authenticated user.
pub struct User {
    pub account_id: u32,
}

#[axum::async_trait]
impl<B> FromRequest<B> for User
where
    B: Send,
{
    type Rejection = http::StatusCode;

    async fn from_request(req: &mut RequestParts<B>) -> Result<Self, Self::Rejection> {
        info!("Authorizing user.");

        let ext = req.extensions().unwrap();

        let database = ext.get::<Arc<Database>>().unwrap();

        let headers = req.headers().unwrap();
        let token = match headers.get(header::AUTHORIZATION.as_str()) {
            None => return Err(http::StatusCode::UNAUTHORIZED),
            Some(header) => header.to_str().unwrap(),
        };

        match sqlx::query!(
            r#"update account_key set last_used = current_timestamp
               where hash = digest($1, 'sha256')
               returning account_id as "account_id!""#,
            token
        )
        .fetch_optional(&database.pool)
        .await
        {
            Err(err) => {
                error!(%err, "Failed to query database");
                Err(http::StatusCode::INTERNAL_SERVER_ERROR)
            }
            Ok(None) => Err(http::StatusCode::UNAUTHORIZED),
            Ok(Some(row)) => Ok(User {
                account_id: row.account_id as u32,
            }),
        }
    }
}

/// Use the acquired OAuth code to retrieve a username from the given
/// provider, then create a new login token for that user.
pub async fn create_token(
    code: String,
    provider_id: &str,
    provider: &Provider,
    database: &Database,
) -> Result<String, String> {
    info!("creating token");

    let url = provider
        .url
        .join("access_token")
        .map_err(|_| "couldn't construct URL".to_string())?;
    let body = serde_urlencoded::to_string(&[
        ("grant_type", "authorization_code"),
        ("code", &code),
        ("redirect_uri", &provider.redirect_url),
    ])
    .unwrap();
    info!("requesting token");

    let response = reqwest::Client::default()
        .post(url)
        .header(reqwest::header::ACCEPT, "application/json")
        .header(
            reqwest::header::CONTENT_TYPE,
            "application/x-www-form-urlencoded",
        )
        .basic_auth(&provider.client_id, Some(&provider.client_secret))
        .body(body)
        .send()
        .await
        .unwrap();
    info!(?response, "received response");

    #[derive(Debug, Deserialize)]
    struct TokenInfo {
        access_token: String,
        // and more fields that we don't need.
    }

    let token_info: TokenInfo = response.json().await.unwrap();
    info!("received token info");

    let mut request = reqwest::Client::default()
        .get(provider.user_url.as_ref())
        .header(reqwest::header::ACCEPT, "application/json")
        .header(
            reqwest::header::AUTHORIZATION,
            format!("Bearer {}", token_info.access_token),
        );
    // GitHub requires a valid user-agent to be set.
    if let Some(ref user_agent) = provider.user_agent {
        request = request.header(reqwest::header::USER_AGENT, user_agent);
    }
    let response = request.send().await.unwrap();
    info!(?response, "received user response");
    if !response.status().is_success() {
        let body = response.text().await.unwrap();
        error!(%body, "failed to get user");
        return Err("couldn't get user".to_string());
    }
    let user_info: serde_json::Value = response.json().await.unwrap();
    info!(%user_info, "decoded user info");

    let username = user_info
        .pointer(&provider.user_path)
        .ok_or_else(|| "couldn't find username field".to_string())?
        .as_str()
        .ok_or_else(|| "username field isn't a string".to_string())?;

    let username = format!("{}:{}", provider_id, username);
    info!(%username, "determined username");

    let account_id = {
        let mut trans = database.pool.begin().await.unwrap();
        let id = match sqlx::query!(
            "select account_id from account where username = $1",
            username
        )
        .fetch_optional(&mut trans)
        .await
        .unwrap()
        {
            Some(row) => row.account_id,
            None => {
                sqlx::query!(
                    "insert into account (username) values ($1) returning account_id",
                    username
                )
                .fetch_one(&mut trans)
                .await
                .unwrap()
                .account_id
            }
        };
        trans.commit().await.unwrap();
        id
    };

    info!(%username, %account_id, "created or retrieved user");

    let token = token(32);
    sqlx::query!(
        "insert into account_key (account_id, hash) values ($1, digest($2, 'sha256'))",
        account_id,
        token
    )
    .execute(&database.pool)
    .await
    .unwrap();

    Ok(token)
}

fn token(len: usize) -> String {
    use rand::Rng;
    rand::thread_rng()
        .sample_iter(&rand::distributions::Alphanumeric)
        .take(len)
        .map(char::from)
        .collect()
}
