//! Web frontend for the crates registry.
//!
//! The frontend is minimal, basic HTML, generated from Handlebars
//! templates.
//!
//! Templates are loaded by default from the `./templates` directory
//! relative to the crate root. This is convenient for development. If
//! the `bundle-templates` feature is enabled, then the templates will
//! be bundled into the binary and unpacked into a tempdir.

use crate::{auth, database::Database};
use axum::{
    extract::{Extension, Path, Query},
    response::{Html,  Redirect},
    routing::get,
};
use handlebars::Handlebars;
use serde::{Deserialize, Serialize};
use std::{collections::BTreeMap, sync::Arc};
use tracing::warn;

/// Unpack template files into a temporary directory. The `TempDir`
/// must not be dropped in order for the templates to remain
/// available.
#[cfg(feature = "bundle-templates")]
fn unpack() -> Result<tempfile::TempDir, std::io::Error> {
    static TEMPLATES: include_dir::Dir = include_dir::include_dir!("./templates");

    let template_dir = tempfile::Builder::new()
        .prefix("crates-templates")
        .tempdir()?;

    TEMPLATES.extract(template_dir.path())?;

    Ok(template_dir)
}

#[derive(Serialize)]
struct Context {
    title: &'static str,
    parent: &'static str,
}

pub fn routes() -> axum::Router {
    let mut handlebars = Handlebars::new();

    #[cfg(feature = "bundle-templates")]
    let tempdir = unpack().unwrap();
    #[cfg(feature = "bundle-templates")]
    let path = tempdir.path();
    #[cfg(not(feature = "bundle-templates"))]
    let path = "templates";

    handlebars
        .register_templates_directory(".hbs", path)
        .unwrap();

    let router = axum::Router::new()
        .route("/", get(index))
        .route("/login", get(login_code))
        .route("/login/:provider_id", get(login_provider))
        .layer(axum::AddExtensionLayer::new(Arc::new(handlebars)));

    #[cfg(feature = "bundle-templates")]
    let router = router.layer(axum::AddExtensionLayer::new(Arc::new(tempdir)));

    router
}

async fn index(handlebars: Extension<Arc<Handlebars<'_>>>) -> Html<String> {
    Html(
        handlebars
            .render(
                "index",
                &Context {
                    title: "Home",
                    parent: "layout",
                },
            )
            .unwrap(),
    )
}

#[derive(Serialize)]
struct LoginContext {
    title: &'static str,
    providers: Vec<LoginProvider>,
    parent: &'static str,
}

#[derive(Serialize)]
struct LoginProvider {
    name: String,
    url: String,
}

async fn login(
    handlebars: Extension<Arc<Handlebars<'_>>>,
    providers: Extension<Arc<BTreeMap<String, auth::Provider>>>,
) -> Html<String> {
    let providers = providers
        .iter()
        .map(|(provider_id, provider)| LoginProvider {
            name: provider.name.clone(),
            url: format!("/login/{}", provider_id),
        })
        .collect();

    Html(
        handlebars
            .render(
                "login",
                &LoginContext {
                    title: "Login",
                    providers,
                    parent: "layout",
                },
            )
            .unwrap(),
    )
}

#[derive(Deserialize)]
struct LoginQuery {
    code: String,
    state: String,
}

#[derive(Serialize)]
struct ApiKeyContext {
    title: &'static str,
    token: String,
    parent: &'static str,
}

async fn login_code(
    handlebars: Extension<Arc<Handlebars<'_>>>,
    query: Option<Query<LoginQuery>>,
    providers: Extension<Arc<BTreeMap<String, auth::Provider>>>,
    database: Extension<Arc<Database>>,
) -> Html<String> {
    if let Some(Query(LoginQuery { code, state })) = query {
        let token = async {
            let provider_id = state
                .split(":")
                .next()
                .expect("str::split is guaranteed to yield at least one element");

            if let Some(provider) = providers.get(provider_id) {
                match crate::auth::create_token(code, provider_id, provider, &database).await {
                    Err(err) => {
                        warn!(%err, "Failed to create auth token");
                        None
                    }
                    Ok(token) => Some(token),
                }
            } else {
                warn!(%state, "Couldn't find OAuth provider");
                None
            }
        };

        match token.await {
            Some(token) => Html(
                handlebars
                    .render(
                        "api_token",
                        &ApiKeyContext {
                            title: "Login",
                            token,
                            parent: "layout",
                        },
                    )
                    .unwrap(),
            ),
            None => login(handlebars, providers).await,
        }
    } else {
        login(handlebars, providers).await
    }
}

async fn login_provider(
    Path(provider_id): Path<String>,
    providers: Extension<Arc<BTreeMap<String, auth::Provider>>>,
) -> Redirect {
    let provider = match providers.get(&provider_id) {
        Some(provider) => provider,
        None => return Redirect::to("/login".parse().unwrap()),
    };

    let state = format!("{}:{}", provider_id, token(8));

    let mut url = provider.url.join("authorize").unwrap();
    url.query_pairs_mut()
        .append_pair("client_id", &provider.client_id)
        .append_pair("redirect_uri", &provider.redirect_url)
        .append_pair("scope", &provider.scope)
        .append_pair("response_type", "code")
        .append_pair("state", &state);
    Redirect::to(url.as_str().parse().unwrap())
}

fn token(len: usize) -> String {
    use rand::Rng;
    rand::thread_rng()
        .sample_iter(&rand::distributions::Alphanumeric)
        .take(len)
        .map(char::from)
        .collect()
}
