// This is an application; private items can be documented.
#![allow(private_intra_doc_links)]

//! This is a Cargo package registry.
//!
//! A minimally functional Cargo registry is fairly simple. This
//! program is based on the [Registries] chapter in the Cargo book.
//!
//! The application serves an HTTP [API][api] and a baisc web
//! [frontend].
//!
//! The registry maintains state in three places:
//!
//! - An [index], stored in a Git repository.
//! - Crate [storage], backed by a filesystem.
//! - A [database] for user management.
//!
//! Configuration is provided using the [`figment`] library. The
//! application looks first for a file called `config.toml` in the
//! current directory, and then for environment variables. See the
//! [`Config`] struct for details.
//!
//! [Registries]: https://doc.rust-lang.org/cargo/reference/registries.html

use axum::AddExtensionLayer;
use serde::Deserialize;
use sqlx::{
    postgres::{PgConnectOptions, PgPoolOptions},
    ConnectOptions,
};
use std::{
    collections::BTreeMap,
    net::IpAddr,
    path::PathBuf,
    sync::{Arc, Mutex},
};
use tower_http::trace::TraceLayer;

mod api;
mod auth;
mod database;
mod de;
mod frontend;
mod index;
mod metrics;
mod storage;

type Result<T> = std::result::Result<T, Error>;

#[derive(Debug, thiserror::Error)]
enum Error {
    #[error("IO: {0}")]
    IO(#[from] std::io::Error),

    #[error("Configuration: {0}")]
    Figment(#[from] figment::Error),

    #[error("Hyper: {0}")]
    Hyper(#[from] hyper::Error),

    #[error("Index: {0}")]
    Index(#[from] index::Error),

    #[error("SQL: {0}")]
    Sql(#[from] sqlx::Error),

    #[error("Migrations: {0}")]
    Migrate(#[from] sqlx::migrate::MigrateError),
}

/// Application configuration.
#[derive(Deserialize)]
struct Config {
    /// The address to bind to.
    address: IpAddr,

    /// The port to bind to.
    port: u16,

    /// The directory where `.crate` files are stored.
    storage_path: PathBuf,

    /// Index repository configuration.
    index: IndexConfig,

    /// Database connection string.
    database_url: String,

    /// A map of OAuth providers that can be used to create users and
    /// generate API tokens. Aside from creating tokens, these OAuth
    /// providers are not used for anything else.
    ///
    /// The application must typically be registered with the OAuth
    /// provider. An example configuration for using GitHub as the
    /// identity provider looks like this:
    ///
    /// ```toml
    /// [auth.github]
    ///   name = "GitHub"
    ///   redirect_url = "http://localhost:4500/login"
    ///   client_id = "..."
    ///   client_secret = "..."
    ///   user_url = "https://api.github.com/user"
    ///   user_path = "/login"
    ///   user_agent = "crates"
    /// ```
    #[serde(default)]
    auth: BTreeMap<String, auth::Provider>,
}

/// Index repository configuration.
#[derive(Deserialize)]
struct IndexConfig {
    /// The path to a Git repository. This could be in the filesystem,
    /// such as `file:///tmp/crates/index.git` or accessible on a
    /// server, such as
    /// `https://git.deschamps.network/crates/index.git`.
    path: String,

    /// If required, an SSH private key for authentication to the
    /// repository.
    private_key: Option<String>,

    /// An optional username, password pair.
    basic_auth: Option<(String, String)>,

    /// The name used for committing to the index.
    name: Option<String>,

    /// The email used for committing to the index.
    email: Option<String>,
}

#[tokio::main]
async fn main() -> Result<()> {
    tracing_subscriber::fmt()
        .with_env_filter(tracing_subscriber::EnvFilter::from_default_env())
        .with_writer(std::io::stderr)
        .init();

    use figment::providers::Format;

    let figment = figment::Figment::from(figment::providers::Serialized::defaults(
        serde_json::json!({
            "address": "127.0.0.1",
            "port": 8000,
        }),
    ))
    .merge(figment::providers::Toml::file("config.toml"))
    .merge(figment::providers::Env::raw());
    let config: Config = figment.extract()?;

    let pool = PgPoolOptions::new()
        .max_connections(4)
        .connect_with({
            let mut options: PgConnectOptions = config.database_url.parse()?;
            options.disable_statement_logging();
            options
        })
        .await?;
    sqlx::migrate!().run(&pool).await?;
    let db = database::Database::new(pool);

    let index = index::Index::new(
        &config.index.path,
        config.index.private_key,
        config.index.basic_auth,
        config.index.name,
        config.index.email,
    )?;

    let storage = storage::Storage {
        path: config.storage_path,
    };

    let app = axum::Router::new()
        .merge(frontend::routes())
        .merge(api::root_routes())
        .merge(metrics::route())
        .nest("/api/v1", api::routes())
        .layer(tower::layer::layer_fn(|inner| metrics::Metrics { inner }))
        .layer(AddExtensionLayer::new(Arc::new(db)))
        .layer(AddExtensionLayer::new(Arc::new(config.auth)))
        .layer(AddExtensionLayer::new(Arc::new(Mutex::new(index))))
        .layer(AddExtensionLayer::new(Arc::new(storage)))
        .layer(TraceLayer::new_for_http());

    axum::Server::bind(&(config.address, config.port).into())
        .serve(app.into_make_service())
        .await?;

    Ok(())
}
