//! Metrics for Prometheus.
//!
//! This module defines various metrics and exposes a [`Service`] that
//! tracks metrics, as well as an [`axum::Router`] with a `/metrics`
//! endpoint.
//!
//! ```
//! let app = axum::Router::new()
//!     .merge(metrics::route())
//!     .layer(tower::layer::layer_fn(|inner| metrics::Metrics { inner }));
//! ```

use axum::{response::Html, routing::get};
use http::{Method, Request};
use prometheus::register_histogram_vec;
use prometheus_static_metric::*;
use std::task::{Context, Poll};
use tower::Service;

make_static_metric! {
    pub struct HttpRequestDuration: Histogram {
        "method" => { get, put, delete, other }
    }
}

lazy_static::lazy_static! {
    /// The amount of time spent handling HTTP requests.
    pub static ref HTTP_REQUEST_DURATION: HttpRequestDuration = register_static_histogram_vec!(
        HttpRequestDuration,
        "http_request_duration",
        "Duration of time spent processing requests.",
        &["method"],
    ).unwrap();
}

pub fn route() -> axum::Router {
    axum::Router::new().route("/metrics", get(metrics))
}

/// A [`Service`] that times requests to include in a `/metrics`
/// endpoint for Prometheus.
#[derive(Clone)]
pub struct Metrics<S> {
    pub inner: S,
}

impl<S, B> Service<Request<B>> for Metrics<S>
where
    S: Service<Request<B>>,
{
    type Response = S::Response;
    type Error = S::Error;
    type Future = S::Future;

    fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
        self.inner.poll_ready(cx)
    }

    fn call(&mut self, mut req: Request<B>) -> Self::Future {
        let histogram = match *req.method() {
            Method::GET => &HTTP_REQUEST_DURATION.get,
            Method::PUT => &HTTP_REQUEST_DURATION.put,
            Method::DELETE => &HTTP_REQUEST_DURATION.delete,
            _ => &HTTP_REQUEST_DURATION.other,
        };
        req.extensions_mut().insert(histogram.start_timer());
        self.inner.call(req)
    }
}

async fn metrics() -> Html<String> {
    use prometheus::Encoder;
    let mut buffer = Vec::new();
    let encoder = prometheus::TextEncoder::new();
    let metric_families = prometheus::gather();
    encoder.encode(&metric_families, &mut buffer).unwrap();
    Html(String::from_utf8(buffer).unwrap())
}
