//! Set up and tear down the test system.
//!
//! This module uses static initialization to start the registry
//! binary and set up its dependencies:
//!
//! - Postgres database
//! - Registry index
//! - Local crate storage directory
//! - Mock auth server
//! - Working directory to hold server config
//! - Local cargo config and `CARGO_HOME`
//!
//! Since the `CARGO_HOME` environment variable is set, the `cargo`
//! binary can be used without specifying the registry; all commands
//! will use the test registry.

mod oauth;

use nix::{sys::signal, unistd::Pid};
use run_script::{run_script, ScriptOptions};
use std::{
    io::Write,
    path::Path,
    process::{Child, Command},
};
use tempfile::TempDir;

/// Similar to lazy_static, but it will be dropped when the process
/// exits. This allows it to shut down the database and server
/// processes and clean up tempdirs.
#[static_init::dynamic(drop)]
static mut TEST_SYSTEM: TestSystem = TestSystem::new();

/// State related to the server being tested and its dependencies.
pub struct TestSystem {
    #[allow(unused)] // Cleans up on drop.
    db_dir: TempDir,
    db_process: Child,

    #[allow(unused)] // Cleans up on drop.
    index_dir: TempDir,

    #[allow(unused)] // Cleans up on drop.
    storage_dir: TempDir,

    #[allow(unused)] // Cleans up on drop.
    auth_server_runtime: tokio::runtime::Runtime,

    #[allow(unused)] // Cleans up on drop.
    server_dir: TempDir,
    server_process: Child,
}

impl TestSystem {
    fn new() -> TestSystem {
        let (db_dir, db_process) = launch_database();

        std::env::set_var("PORT", "4500");
        let index_dir = create_index();

        let storage_dir = tempfile::Builder::new()
            .prefix("crates-storage")
            .tempdir()
            .unwrap();

        let auth_server_port = 4501;
        let auth_server_runtime = oauth::mock_server(auth_server_port);

        let (server_dir, server_process) =
            launch_server(index_dir.path(), storage_dir.path(), auth_server_port);

        TestSystem {
            db_dir,
            db_process,
            index_dir,
            storage_dir,
            auth_server_runtime,
            server_dir,
            server_process,
        }
    }
}

fn sigint(child: &mut Child) {
    signal::kill(Pid::from_raw(child.id() as i32), signal::Signal::SIGINT).unwrap();
    child.wait().unwrap();
}

impl Drop for TestSystem {
    fn drop(&mut self) {
        sigint(&mut self.server_process);
        sigint(&mut self.db_process);
    }
}

fn launch_database() -> (TempDir, Child) {
    let data_dir = tempfile::Builder::new()
        .prefix("crates-db")
        .tempdir()
        .unwrap();

    let pgdata = data_dir.path();

    std::env::set_var("PGDATA", pgdata);
    std::env::set_var("PGHOST", pgdata);
    std::env::set_var(
        "DATABASE_URL",
        &format!("postgres:///crates?host={}", pgdata.display()),
    );

    let (code, _, _) = run_script!(
        r#"
        initdb
        # NOTE: There are tabs in here, because heredoc requires them.
        cat >> $PGDATA/postgresql.conf <<-EOF
			listen_addresses = ''
			unix_socket_directories = '$PGHOST'
		EOF
        echo "CREATE DATABASE crates;" | postgres --single -E postgres
        "#
    )
    .unwrap();
    assert_eq!(code, 0);

    let child = Command::new("postgres").spawn().unwrap();

    // Wait long enough for postgres to start up.
    std::thread::sleep(std::time::Duration::from_secs(1));

    let (code, _, _) = run_script!("sqlx migrate run").unwrap();
    assert_eq!(code, 0);

    (data_dir, child)
}

/// Create and initialize a new index repo.
fn create_index() -> TempDir {
    let index_dir = tempfile::Builder::new()
        .prefix("crates-index")
        .tempdir()
        .unwrap();

    let mut options = run_script::ScriptOptions::new();
    options.working_directory = Some(index_dir.path().into());

    // Initialize the index. It needs to be a bare repository so that
    // the server can push to it, but we need to clone it in order to
    // have a worktree from which to initialize it.
    let (code, _, _) = run_script!(
        r#"
        git init --bare -b master
        mkdir checkout
        cd checkout
        git clone .. .
        # NOTE: There are tabs in here, because heredoc requires them.
		cat >> config.json <<-EOF
			{
			    "dl": "http://localhost:$PORT/api/v1/crates",
			    "api": "http://localhost:$PORT"
			}
		EOF
        git add config.json
        git commit -m "Create index"
        git push
        "#,
        &options
    )
    .unwrap();
    assert_eq!(code, 0);

    index_dir
}

fn launch_server(index: &Path, storage: &Path, auth_port: u16) -> (TempDir, Child) {
    let (code, secret_key, _) = run_script!("openssl rand -base64 32").unwrap();
    assert_eq!(code, 0);
    std::env::set_var("SECRET_KEY", secret_key);

    let config = format!(
        r#"
storage_path = "{storage}"

[index]
  path = "{index}"

[auth.mock]
  name = "Mock Auth Provider"
  url = "http://localhost:{auth_port}/login/oauth/"
  redirect_url = "http://localhost:{port}/login"
  client_id = "mockclientid"
  client_secret = "mockclientsecret"
  user_url = "http://localhost:{auth_port}/user"
  user_path = "/login"
"#,
        port = 4500,
        storage = storage.display(),
        index = index.display(),
        auth_port = auth_port
    );

    let server_dir = tempfile::Builder::new()
        .prefix("crates-server")
        .tempdir()
        .unwrap();
    let mut config_file = std::fs::File::create(server_dir.path().join("config.toml")).unwrap();
    config_file.write_all(config.as_bytes()).unwrap();

    let server_bin = std::path::Path::new("./target/debug/cargo-registry")
        .canonicalize()
        .unwrap();

    let child = Command::new(server_bin)
        .current_dir(server_dir.path())
        .spawn()
        .unwrap();

    // Wait long enough for the server to start up.
    std::thread::sleep(std::time::Duration::from_secs(1));

    (server_dir, child)
}

pub struct Cargo {
    #[allow(unused)]
    home_dir: TempDir,
    options: ScriptOptions,
}

impl Cargo {
    pub fn new() -> Self {
        let home_dir = tempfile::Builder::new()
            .prefix("crates-cargo")
            .tempdir()
            .unwrap();

        let config = format!(
            r#"
[registry]
default = "test"

[registries]
test = {{ index = "file://{index}" }}
"#,
            index = TEST_SYSTEM.read().index_dir.path().display()
        );

        let mut config_file = std::fs::File::create(home_dir.path().join("config")).unwrap();
        config_file.write_all(config.as_bytes()).unwrap();

        let mut options = ScriptOptions::new();
        options.working_directory = Some(home_dir.path().into());
        let (code, _, _) = run_script!(
            format!(
                r#"
                cat >> run.sh <<-EOF
					#!/bin/sh
					export CARGO_HOME={}
					sh -e "\$@"
				EOF
                "#,
                home_dir.path().display()
            ),
            &options
        )
        .unwrap();
        assert_eq!(code, 0);
        let (code, _, _) = run_script!("chmod +x run.sh", &options).unwrap();
        assert_eq!(code, 0);
        options.runner = Some(format!("{}/run.sh", home_dir.path().display()));

        Cargo { home_dir, options }
    }

    /// Get a `ScriptOptions` object that is configured to run scripts
    /// with `CARGO_HOME` set to this Cargo instance.
    pub fn script_options(&self) -> ScriptOptions {
        self.options.clone()
    }
}
