//! Mock OAuth provider.

use axum::{
    extract::{Extension, Form, Query},
    response::Redirect,
    routing::{get, post},
    Json,
};
use serde::{Deserialize, Serialize};
use std::sync::{
    atomic::{AtomicUsize, Ordering},
    Arc,
};

struct UserCounter(AtomicUsize);

pub fn mock_server(port: u16) -> tokio::runtime::Runtime {
    let runtime = tokio::runtime::Builder::new_multi_thread()
        .thread_name("oauth-rt")
        .enable_all()
        .build()
        .unwrap();

    let user_counter = UserCounter(AtomicUsize::new(0));

    let app = axum::Router::new()
        .route("/login/oauth/authorize", get(authorize))
        .route("/login/oauth/access_token", post(access_token))
        .route("/user", get(user))
        .layer(axum::AddExtensionLayer::new(Arc::new(user_counter)));

    runtime.spawn(async move {
        axum::Server::bind(&format!("127.0.0.1:{}", port).parse().unwrap())
            .serve(app.into_make_service())
            .await
            .unwrap();
    });

    runtime
}

#[derive(Deserialize)]
struct AuthorizeQuery {
    #[allow(unused)]
    client_id: String,
    redirect_uri: String,
    #[allow(unused)]
    scope: String,
    #[allow(unused)]
    response_type: String,
    state: String,
}

async fn authorize(query: Query<AuthorizeQuery>) -> Redirect {
    Redirect::to(
        format!("{}?code=foo&state={}", query.redirect_uri, query.state)
            .parse()
            .unwrap(),
    )
}

#[derive(Deserialize)]
struct AccessTokenData {
    #[allow(unused)]
    grant_type: String,
    #[allow(unused)]
    code: String,
    #[allow(unused)]
    redirect_uri: String,
}

#[derive(Debug, Serialize)]
struct TokenInfo {
    access_token: String,
    // and more fields that we don't need.
}

async fn access_token(#[allow(unused)] data: Form<AccessTokenData>) -> Json<TokenInfo> {
    let access_token = "mocktoken".to_string();
    Json(TokenInfo { access_token })
}

#[derive(Debug, Serialize)]
struct User {
    login: String,
}

async fn user(Extension(counter): Extension<Arc<UserCounter>>) -> Json<User> {
    let user_number = counter.0.fetch_add(1, Ordering::SeqCst);
    let login = format!("user-{}", user_number);
    Json(User { login })
}
