// This module statically sets up and tears down the test fixture.
mod fixture;

use run_script::run_script;
use tempfile::TempDir;
use testdep::testdep;

/// Run a script with the given script options, assert that the exit
/// status is 0, and return stdout.
macro_rules! script {
    ($options:expr, $script:expr) => {{
        let (code, stdout, _) = dbg!(run_script!($script, &$options).unwrap());
        assert_eq!(code, 0);
        stdout
    }};
}

#[test]
fn search_empty_results() {
    let cargo = fixture::Cargo::new();
    let options = cargo.script_options();
    let output = script!(options, "cargo search a-crate-that-does-not-exist");
    assert_eq!(output, "");
}

#[testdep]
fn create_account() -> fixture::Cargo {
    let cargo = fixture::Cargo::new();
    let options = cargo.script_options();

    let output = script!(options, "curl -s -L http://localhost:4500/login/mock");
    let re = regex::Regex::new(r"Your API token is <b>(.+)</b>").unwrap();
    let caps = re.captures(&output).unwrap();
    let token = caps.get(1).unwrap().as_str();
    script!(options, format!("cargo login --registry test {}", token));

    // Check the username. The mock OAuth provider generates usernames
    // with incrementing numbers.
    let output = script!(
        options,
        format!(
            "curl -s -H 'Authorization: {}' http://localhost:4500/me | jq -j '.username'",
            token
        )
    );
    assert_eq!(output, "mock:user-0");

    cargo
}

#[testdep]
fn create_second_account(
    // Force the other test to complete first to ensure that accounts
    // are named `mock:user-0` and `mock:user-1` respectively.
    #[allow(unused)] create_account: &fixture::Cargo,
) -> fixture::Cargo {
    let cargo = fixture::Cargo::new();
    let options = cargo.script_options();

    let output = script!(options, "curl -s -L http://localhost:4500/login/mock");
    let re = regex::Regex::new(r"Your API token is <b>(.+)</b>").unwrap();
    let caps = re.captures(&output).unwrap();
    let token = caps.get(1).unwrap().as_str();
    script!(options, format!("cargo login --registry test {}", token));

    let output = script!(
        options,
        format!(
            "curl -s -H 'Authorization: {}' http://localhost:4500/me | jq -j '.username'",
            token
        )
    );
    assert_eq!(output, "mock:user-1");

    cargo
}

#[testdep]
fn publish_crate(create_account: &fixture::Cargo) -> TempDir {
    let mut options = create_account.script_options();
    let crate_dir = tempfile::Builder::new()
        .prefix("crates-test")
        .tempdir()
        .unwrap();
    options.working_directory = Some(crate_dir.path().into());

    script!(options, "cargo init --lib --name testcrate");
    script!(
        options,
        r#"
        cat >> src/lib.rs <<-EOF
			pub fn version() -> String {
				format!(
					"{} version {}",
					env!("CARGO_PKG_NAME"),
					env!("CARGO_PKG_VERSION"),
				)
			}
		EOF
        "#
    );
    script!(options, "cargo publish --allow-dirty");

    let output = script!(options, "cargo search testcrate");
    let re = regex::Regex::new(r#"testcrate = "0\.1\.0""#).unwrap();
    assert!(re.is_match(&output));

    crate_dir
}

#[testdep]
fn add_crate_owner(
    create_account: &fixture::Cargo,
    #[allow(unused)] create_second_account: &fixture::Cargo,
    publish_crate: &TempDir,
) {
    let mut options = create_account.script_options();
    options.working_directory = Some(publish_crate.path().into());

    let re = regex::Regex::new(r#"mock:user-1"#).unwrap();
    let output = script!(options, "cargo owner --list");
    assert!(!re.is_match(&output));
    script!(options, format!("cargo owner --add {}", "mock:user-1"));
    let output = script!(options, "cargo owner --list");
    assert!(re.is_match(&output));
}

/// Remove the original author from the crate owners.
#[testdep]
fn remove_crate_owner(
    create_second_account: &fixture::Cargo,
    publish_crate: &TempDir,
    #[allow(unused)] update_crate: &(),
) {
    let mut options = create_second_account.script_options();
    options.working_directory = Some(publish_crate.path().into());

    let re = regex::Regex::new(r#"mock:user-0"#).unwrap();
    let output = script!(options, "cargo owner --list");
    assert!(re.is_match(&output));
    script!(options, format!("cargo owner --remove {}", "mock:user-0"));
    let output = script!(options, "cargo owner --list");
    assert!(!re.is_match(&output));
}

#[testdep]
fn update_crate(
    create_second_account: &fixture::Cargo,
    publish_crate: &TempDir,
    // user-1 must be an owner first.
    #[allow(unused)] add_crate_owner: (),
    #[allow(unused)] use_library: &(fixture::Cargo, TempDir),
) {
    let mut options = create_second_account.script_options();
    options.working_directory = Some(publish_crate.path().into());

    script!(options, "sed -i 's/0.1.0/0.1.1/' Cargo.toml");
    script!(options, "cargo publish --allow-dirty");
    let output = script!(options, "cargo search testcrate");
    let re = regex::Regex::new(r#"testcrate = "0\.1\.1""#).unwrap();
    assert!(re.is_match(&output));
}

#[testdep]
fn use_library(#[allow(unused)] publish_crate: &TempDir) -> (fixture::Cargo, TempDir) {
    let cargo = fixture::Cargo::new();
    let mut options = cargo.script_options();
    let dir = tempfile::tempdir().unwrap();
    options.working_directory = Some(dir.path().into());

    script!(options, "cargo init --bin --name testbin");
    script!(
        options,
        r#"
        echo 'testcrate = { "version" = "0.1", registry = "test" }' >> Cargo.toml
        cat > src/main.rs <<-EOF
			fn main() {
				print!("{}", testcrate::version());
			}
		EOF
        "#
    );
    let output = script!(options, "cargo run");
    assert_eq!(output, "testcrate version 0.1.0");

    (cargo, dir)
}

#[testdep]
fn update_dependencies(#[allow(unused)] update_crate: &(), use_library: (fixture::Cargo, TempDir)) {
    let (cargo, dir) = use_library;
    let mut options = cargo.script_options();
    options.working_directory = Some(dir.path().into());

    script!(options, "cargo update");
    let output = script!(options, "cargo run");
    assert_eq!(output, "testcrate version 0.1.1");
}
